from __future__ import print_function
from aiida.orm.calculation.work import WorkCalculation
from aiida.orm.data.base import Float, Str
from aiida.work import workfunction as wf
import sys
import os
import numpy as np
import pandas as pd
from calc_pe.utils import printPE

ParameterData = DataFactory('parameter')
StructureData = DataFactory('structure')
CifData = DataFactory('cif')

@wf
def CalcPE(ResGasCO2, ResGasN2, GasIn, VF, Process, Cp, Yd, ElEff, Opt):
    """ Submit calc_pe calculation using AiiDA, for the CO2 parasitic energy.
    :ParameterData ResGasCO2:
    :ParameterData ResGasN2:
    :Str GasIn:
    :Float VF:
    :Str Process:
    :Float Cp:
    :Float Yd:
    :Str ElEff:
    :Str Opt:
    """
    from calc_pe import mainPE
    T_iso = {}
    iso_df = {}
    for i, ResGas in enumerate([ResGasCO2, ResGasN2]):
        gas = ['CO_2','N_2'][i]
        res = ResGas.get_dict()
        T_iso[gas] = res["temperature"]
        iso_df[gas] = pd.DataFrame(columns = ['pressure(Pa)','loading(mol/kg)','HoA(kJ/mol)'])
        iso_df[gas]['pressure(Pa)'] = [a[0]*1e5 for a in res['isotherm_loading']] # converted from bar to Pa
        iso_df[gas]['loading(mol/kg)'] = [a[1] for a in res['isotherm_loading']]
        iso_df[gas]['HoA(kJ/mol)'] = [a[1] for a in res['isotherm_enthalpy']]
        # TRICK: use the enthalpy from widom (energy-RT) which is more accurate
        #        that the one at 0.001 bar (and which also is NaN for weakly interacting systems)
        iso_df[gas]['HoA(kJ/mol)'].loc[0] = res['adsorption_energy_average']-res['temperature']/120.027

    pe_dict = mainPE(gasin = GasIn.value,        # e.g., "coal"
                      rho = res['Density']*1000, # converted from g/cm3 to kg/m3
                      vf = VF.value,             # e.g., 0.35
                      process = Process.value,   # e.g., "TPSA"
                      cp = Cp.value,             # e.g., 985.0, the average for MOFs in J/kg/K
                      yd = Yd.value,             # e.g., 0.99
                      eleff = ElEff.value,       # e.g., "carnot"
                      opt = Opt.value,           # e.g., "PE"
                      T_iso = T_iso,             # [T_iso_CO2, T_iso_N2]
                      iso_df = iso_df            # [iso_df_CO2, iso_df_N2] df containing pressure (Pa), loading (mol/kg), HeatOfAdsorption (kJ/mol)
                     )
    return ParameterData(dict=pe_dict)

# For every structure, parse co2 and n2 isotherms and submit to calcPE
df = pd.read_csv("cof_ids.csv")

for i in df.index:

    isotherms = {}

    for gas in ['co2','n2']:

	    q = QueryBuilder()
	    q.append(StructureData, 
		     filters={'label':df.at[i,'cof_label']}, 
		     tag='inp_struct'
		     )
	    q.append(WorkCalculation, 
		     filters={'label': df.at[i,'cellopt_label']},
		     output_of='inp_struct', 
		     tag='celloptworkcalc'
                     )
	    q.append(CifData, 
		     edge_filters={'label': 'output_structure'},
		     output_of='celloptworkcalc',
                     tag='optcif',
                     )
	    q.append(WorkCalculation, 
		     filters={'label': 'isotherm-xx'.replace('xx',gas)},
		     output_of='optcif', 
		     tag='isothermworkcalc'
                     )
            q.append(ParameterData, 
                     output_of='isothermworkcalc',
                     tag='isothermpd',
                     )

            if len(q.all())>0:
                isotherms[gas] = q.all()[0][0]

            # Check if the structure is present, is porous (has 'isotherm_loading'), then compute PE
            if ('co2' in isotherms.keys() and
		'n2' in isotherms.keys() and
		'isotherm_loading' in isotherms['co2'].get_dict() and
		'isotherm_loading' in isotherms['n2'].get_dict()):
		ResultPE = CalcPE(ResGasCO2=isotherms['co2'],
		                  ResGasN2=isotherms['n2'],
		                  GasIn=Str('coal'),
		                  VF=Float(0.35),
		                  Process=Str('TPSA'),
		                  Cp=Float(985.0),
		                  Yd=Float(0.99),
		                  ElEff=Str('carnot'),
		                  Opt=Str('PE'),
		                  )
	        printPE(df.at[i,'cof_label'],ResultPE.get_dict())
